@@ -58,12 +58,9 @@ fi
 # build the version we want from source
 if test -v TEST_MOD_TLS; then
   RUSTLS_HOME="$HOME/build/rustls-ffi"
-  RUSTLS_VERSION="v0.9.0"
-  git clone https://github.com/rustls/rustls-ffi.git "$RUSTLS_HOME"
+  RUSTLS_VERSION="v0.10.0"
+  git clone -b "$RUSTLS_VERSION" https://github.com/rustls/rustls-ffi.git "$RUSTLS_HOME"
   pushd "$RUSTLS_HOME"
-    # since v0.9.0, there is no longer a dependency on cbindgen
-    git fetch origin
-    git checkout tags/$RUSTLS_VERSION
     make install DESTDIR="$PREFIX"
   popd
   CONFIG="$CONFIG --with-tls --with-rustls=$PREFIX"
@@ -196,6 +193,8 @@ if ! test -v SKIP_TESTING; then
     fi
 
     if test -v TEST_H2 -a $RV -eq 0; then
+        # Build the test clients
+        (cd test/clients && make)
         # Run HTTP/2 tests.
         MPM=event py.test-3 test/modules/http2
         RV=$?
